@@ -26,7 +26,7 @@
 #ifndef _COMPIZ_H
 #define _COMPIZ_H
 
-#define ABIVERSION 20060927
+#define ABIVERSION 20061112
 
 #include <stdio.h>
 #include <sys/time.h>
@@ -137,15 +137,16 @@ typedef struct _CompIcon    CompIcon;
 #define MAXIMIZE_STATE (CompWindowStateMaximizedHorzMask | \
 			CompWindowStateMaximizedVertMask)
 
-#define CompWindowActionMoveMask	 (1 << 0)
-#define CompWindowActionResizeMask	 (1 << 1)
-#define CompWindowActionStickMask	 (1 << 2)
-#define CompWindowActionMinimizeMask     (1 << 3)
-#define CompWindowActionMaximizeHorzMask (1 << 4)
-#define CompWindowActionMaximizeVertMask (1 << 5)
-#define CompWindowActionFullscreenMask	 (1 << 6)
-#define CompWindowActionCloseMask	 (1 << 7)
-#define CompWindowActionShadeMask	 (1 << 8)
+#define CompWindowActionMoveMask	  (1 << 0)
+#define CompWindowActionResizeMask	  (1 << 1)
+#define CompWindowActionStickMask	  (1 << 2)
+#define CompWindowActionMinimizeMask      (1 << 3)
+#define CompWindowActionMaximizeHorzMask  (1 << 4)
+#define CompWindowActionMaximizeVertMask  (1 << 5)
+#define CompWindowActionFullscreenMask	  (1 << 6)
+#define CompWindowActionCloseMask	  (1 << 7)
+#define CompWindowActionShadeMask	  (1 << 8)
+#define CompWindowActionChangeDesktopMask (1 << 9)
 
 #define MwmFuncAll      (1L << 0)
 #define MwmFuncResize   (1L << 1)
@@ -203,6 +204,7 @@ extern Bool       replaceCurrentWm;
 extern Bool       indirectRendering;
 extern Bool       strictBinding;
 extern Bool       useCow;
+extern Bool       noDetection;
 
 extern int  defaultRefreshRate;
 extern char *defaultTextureFilter;
@@ -291,9 +293,10 @@ typedef enum {
 } CompOptionType;
 
 typedef enum {
-    CompBindingTypeNone   = 0,
-    CompBindingTypeKey    = 1 << 0,
-    CompBindingTypeButton = 1 << 1
+    CompBindingTypeNone       = 0,
+    CompBindingTypeKey        = 1 << 0,
+    CompBindingTypeButton     = 1 << 1,
+    CompBindingTypeEdgeButton = 1 << 2
 } CompBindingType;
 
 typedef enum {
@@ -344,6 +347,7 @@ struct _CompAction {
     Bool bell;
 
     unsigned int edgeMask;
+    int		 edgeButton;
 };
 
 typedef struct {
@@ -457,6 +461,40 @@ getStringOptionNamed (CompOption *option,
 		      char	 *name,
 		      char	 *defaultValue);
 
+unsigned short *
+getColorOptionNamed (CompOption	    *option,
+		     int	    nOption,
+		     char	    *name,
+		     unsigned short *defaultValue);
+
+char *
+keyBindingToString (CompDisplay    *d,
+		    CompKeyBinding *key);
+
+char *
+buttonBindingToString (CompDisplay       *d,
+		       CompButtonBinding *button);
+
+Bool
+stringToKeyBinding (CompDisplay    *d,
+		    const char     *binding,
+		    CompKeyBinding *key);
+
+Bool
+stringToButtonBinding (CompDisplay	 *d,
+		       const char	 *binding,
+		       CompButtonBinding *button);
+
+char *
+edgeToString (unsigned int edge);
+
+Bool
+stringToColor (const char     *color,
+	       unsigned short *rgba);
+
+char *
+colorToString (unsigned short *rgba);
+
 
 /* display.c */
 
@@ -496,28 +534,30 @@ typedef int CompWatchFdHandle;
 #define COMP_DISPLAY_OPTION_RUN_COMMAND10                 30
 #define COMP_DISPLAY_OPTION_RUN_COMMAND11                 31
 #define COMP_DISPLAY_OPTION_SLOW_ANIMATIONS               32
-#define COMP_DISPLAY_OPTION_LOWER_WINDOW                  33
-#define COMP_DISPLAY_OPTION_UNMAXIMIZE_WINDOW             34
-#define COMP_DISPLAY_OPTION_MINIMIZE_WINDOW               35
-#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW               36
-#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_HORZ          37
-#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_VERT          38
-#define COMP_DISPLAY_OPTION_OPACITY_INCREASE              39
-#define COMP_DISPLAY_OPTION_OPACITY_DECREASE              40
-#define COMP_DISPLAY_OPTION_SCREENSHOT                    41
-#define COMP_DISPLAY_OPTION_RUN_SCREENSHOT                42
-#define COMP_DISPLAY_OPTION_WINDOW_SCREENSHOT             43
-#define COMP_DISPLAY_OPTION_RUN_WINDOW_SCREENSHOT         44
-#define COMP_DISPLAY_OPTION_WINDOW_MENU                   45
-#define COMP_DISPLAY_OPTION_SHOW_DESKTOP                  46
-#define COMP_DISPLAY_OPTION_RAISE_ON_CLICK                47
-#define COMP_DISPLAY_OPTION_AUDIBLE_BELL                  48
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED       49
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_HORZ  50
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_VERT  51
-#define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS     52
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED          53
-#define COMP_DISPLAY_OPTION_NUM                           54
+#define COMP_DISPLAY_OPTION_RAISE_WINDOW                  33
+#define COMP_DISPLAY_OPTION_LOWER_WINDOW                  34
+#define COMP_DISPLAY_OPTION_UNMAXIMIZE_WINDOW             35
+#define COMP_DISPLAY_OPTION_MINIMIZE_WINDOW               36
+#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW               37
+#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_HORZ          38
+#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_VERT          39
+#define COMP_DISPLAY_OPTION_OPACITY_INCREASE              40
+#define COMP_DISPLAY_OPTION_OPACITY_DECREASE              41
+#define COMP_DISPLAY_OPTION_SCREENSHOT                    42
+#define COMP_DISPLAY_OPTION_RUN_SCREENSHOT                43
+#define COMP_DISPLAY_OPTION_WINDOW_SCREENSHOT             44
+#define COMP_DISPLAY_OPTION_RUN_WINDOW_SCREENSHOT         45
+#define COMP_DISPLAY_OPTION_WINDOW_MENU                   46
+#define COMP_DISPLAY_OPTION_SHOW_DESKTOP                  47
+#define COMP_DISPLAY_OPTION_RAISE_ON_CLICK                48
+#define COMP_DISPLAY_OPTION_AUDIBLE_BELL                  49
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED       50
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_HORZ  51
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_VERT  52
+#define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS     53
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED          54
+#define COMP_DISPLAY_OPTION_IGNORE_HINTS_WHEN_MAXIMIZED   55
+#define COMP_DISPLAY_OPTION_NUM                           56
 
 typedef CompOption *(*GetDisplayOptionsProc) (CompDisplay *display,
 					      int	  *count);
@@ -538,6 +578,12 @@ typedef void (*FiniPluginForDisplayProc) (CompPlugin  *plugin,
 typedef void (*HandleEventProc) (CompDisplay *display,
 				 XEvent	     *event);
 
+typedef void (*HandleCompizEventProc) (CompDisplay *display,
+				       char        *pluginName,
+				       char        *eventName,
+				       CompOption  *option,
+				       int         nOption);
+
 typedef Bool (*CallBackProc) (void *closure);
 
 typedef void (*ForEachWindowProc) (CompWindow *window,
@@ -629,6 +675,7 @@ struct _CompDisplay {
     Atom winActionFullscreenAtom;
     Atom winActionCloseAtom;
     Atom winActionShadeAtom;
+    Atom winActionChangeDesktopAtom;
 
     Atom wmAllowedActionsAtom;
 
@@ -719,7 +766,8 @@ struct _CompDisplay {
     InitPluginForDisplayProc initPluginForDisplay;
     FiniPluginForDisplayProc finiPluginForDisplay;
 
-    HandleEventProc handleEvent;
+    HandleEventProc       handleEvent;
+    HandleCompizEventProc handleCompizEvent;
 
     CompPrivate *privates;
 };
@@ -765,6 +813,9 @@ addDisplay (char *name,
 	    char **plugin,
 	    int  nPlugin);
 
+Time
+getCurrentTimeFromDisplay (CompDisplay *d);
+
 void
 focusDefaultWindow (CompDisplay *d);
 
@@ -824,6 +875,13 @@ void
 handleEvent (CompDisplay *display,
 	     XEvent      *event);
 
+void
+handleCompizEvent (CompDisplay *display,
+		   char        *pluginName,
+		   char        *eventName,
+		   CompOption  *option,
+		   int         nOption);
+
 void
 handleSyncAlarm (CompWindow *w);
 
@@ -837,6 +895,10 @@ eventTerminates (CompDisplay *display,
 		 XEvent      *event,
 		 CompOption  *option);
 
+void
+clearTargetOutput (CompDisplay	*display,
+		   unsigned int mask);
+
 /* paint.c */
 
 #define MULTIPLY_USHORT(us1, us2)		 \
@@ -900,9 +962,14 @@ typedef Bool (*PaintScreenProc) (CompScreen		 *screen,
 
 typedef void (*PaintTransformedScreenProc) (CompScreen		    *screen,
 					    const ScreenPaintAttrib *sAttrib,
+					    Region		    region,
 					    int			    output,
 					    unsigned int	    mask);
 
+typedef void (*ApplyScreenTransformProc) (CompScreen		  *screen,
+					  const ScreenPaintAttrib *sAttrib,
+					  int			  output);
+
 
 #define PAINT_WINDOW_SOLID_MASK			(1 << 0)
 #define PAINT_WINDOW_TRANSLUCENT_MASK		(1 << 1)
@@ -945,12 +1012,23 @@ donePaintScreen (CompScreen *screen);
 void
 translateRotateScreen (const ScreenPaintAttrib *sa);
 
+void
+prepareXCoords (CompScreen *screen,
+		int	   output,
+		float      z);
+
 void
 paintTransformedScreen (CompScreen		*screen,
 			const ScreenPaintAttrib *sAttrib,
+			Region			region,
 			int			output,
 			unsigned int	        mask);
 
+void
+applyScreenTransform (CompScreen	      *screen,
+		      const ScreenPaintAttrib *sAttrib,
+		      int		      output);
+
 Bool
 paintScreen (CompScreen		     *screen,
 	     const ScreenPaintAttrib *sAttrib,
@@ -1013,6 +1091,7 @@ struct _CompTexture {
     CompMatrix matrix;
     Bool       oldMipmaps;
     Bool       mipmap;
+    int        refCount;
 };
 
 void
@@ -1023,6 +1102,13 @@ void
 finiTexture (CompScreen  *screen,
 	     CompTexture *texture);
 
+CompTexture *
+createTexture (CompScreen *screen);
+
+void
+destroyTexture (CompScreen  *screen,
+		CompTexture *texture);
+
 Bool
 readImageToTexture (CompScreen   *screen,
 		    CompTexture  *texture,
@@ -1084,7 +1170,10 @@ disableTexture (CompScreen  *screen,
 #define COMP_SCREEN_OPTION_UNREDIRECT_FS       6
 #define COMP_SCREEN_OPTION_DEFAULT_ICON        7
 #define COMP_SCREEN_OPTION_SYNC_TO_VBLANK      8
-#define COMP_SCREEN_OPTION_NUM		       9
+#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS  9
+#define COMP_SCREEN_OPTION_DETECT_OUTPUTS      10
+#define COMP_SCREEN_OPTION_OUTPUTS	       11
+#define COMP_SCREEN_OPTION_NUM		       12
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -1235,6 +1324,10 @@ typedef void (*WindowGrabNotifyProc) (CompWindow   *window,
 
 typedef void (*WindowUngrabNotifyProc) (CompWindow *window);
 
+typedef void (*WindowStateChangeNotifyProc) (CompWindow *window);
+
+typedef void (*OutputChangeNotifyProc) (CompScreen *screen);
+
 #define COMP_SCREEN_DAMAGE_PENDING_MASK (1 << 0)
 #define COMP_SCREEN_DAMAGE_REGION_MASK  (1 << 1)
 #define COMP_SCREEN_DAMAGE_ALL_MASK     (1 << 2)
@@ -1303,8 +1396,11 @@ struct _CompIcon {
 };
 
 typedef struct _CompOutput {
-    char   *name;
-    REGION region;
+    char       *name;
+    REGION     region;
+    int        width;
+    int        height;
+    XRectangle workArea;
 } CompOutput;
 
 struct _CompScreen {
@@ -1324,6 +1420,8 @@ struct _CompScreen {
     int		      y;
     int		      hsize;		/* Number of horizontal viewports */
     int		      vsize;		/* Number of vertical viewports */
+    unsigned int      nDesktop;
+    unsigned int      currentDesktop;
     REGION	      region;
     Region	      damage;
     unsigned long     damageMask;
@@ -1406,6 +1504,8 @@ struct _CompScreen {
 
     GLint stencilRef;
 
+    Bool clearBuffers;
+
     Bool lighting;
     Bool slowAnimations;
 
@@ -1415,6 +1515,9 @@ struct _CompScreen {
 
     unsigned int showingDesktopMask;
 
+    unsigned long *desktopHintData;
+    int           desktopHintSize;
+
     GLXGetProcAddressProc    getProcAddress;
     GLXBindTexImageProc      bindTexImage;
     GLXReleaseTexImageProc   releaseTexImage;
@@ -1452,24 +1555,29 @@ struct _CompScreen {
     InitPluginForScreenProc initPluginForScreen;
     FiniPluginForScreenProc finiPluginForScreen;
 
-    PreparePaintScreenProc       preparePaintScreen;
-    DonePaintScreenProc	         donePaintScreen;
-    PaintScreenProc	         paintScreen;
-    PaintTransformedScreenProc   paintTransformedScreen;
-    PaintBackgroundProc          paintBackground;
-    PaintWindowProc	         paintWindow;
-    AddWindowGeometryProc        addWindowGeometry;
-    DrawWindowGeometryProc       drawWindowGeometry;
-    DrawWindowTextureProc	 drawWindowTexture;
-    DamageWindowRectProc         damageWindowRect;
-    FocusWindowProc		 focusWindow;
-    SetWindowScaleProc		 setWindowScale;
+    PreparePaintScreenProc     preparePaintScreen;
+    DonePaintScreenProc	       donePaintScreen;
+    PaintScreenProc	       paintScreen;
+    PaintTransformedScreenProc paintTransformedScreen;
+    ApplyScreenTransformProc   applyScreenTransform;
+    PaintBackgroundProc        paintBackground;
+    PaintWindowProc	       paintWindow;
+    AddWindowGeometryProc      addWindowGeometry;
+    DrawWindowGeometryProc     drawWindowGeometry;
+    DrawWindowTextureProc      drawWindowTexture;
+    DamageWindowRectProc       damageWindowRect;
+    FocusWindowProc	       focusWindow;
+    SetWindowScaleProc	       setWindowScale;
 
     WindowResizeNotifyProc windowResizeNotify;
     WindowMoveNotifyProc   windowMoveNotify;
     WindowGrabNotifyProc   windowGrabNotify;
     WindowUngrabNotifyProc windowUngrabNotify;
 
+    WindowStateChangeNotifyProc windowStateChangeNotify;
+
+    OutputChangeNotifyProc outputChangeNotify;
+
     CompPrivate *privates;
 };
 
@@ -1657,6 +1765,37 @@ outputDeviceForPoint (CompScreen *s,
 		      int	 x,
 		      int	 y);
 
+void
+getCurrentOutputExtents (CompScreen *s,
+			 int	    *x1,
+			 int	    *y1,
+			 int	    *x2,
+			 int	    *y2);
+
+void
+getWorkareaForOutput (CompScreen *s,
+		      int	 output,
+		      XRectangle *area);
+
+void
+setNumberOfDesktops (CompScreen   *s,
+		     unsigned int nDesktop);
+
+void
+setCurrentDesktop (CompScreen   *s,
+		   unsigned int desktop);
+
+void
+setDefaultViewport (CompScreen *s);
+
+void
+outputChangeNotify (CompScreen *s);
+
+void
+clearScreenOutput (CompScreen   *s,
+		   int	        output,
+		   unsigned int mask);
+
 
 /* window.c */
 
@@ -1700,11 +1839,14 @@ struct _CompWindow {
     XWindowAttributes attrib;
     int		      serverX;
     int		      serverY;
+    int		      serverWidth;
+    int		      serverHeight;
+    int		      serverBorderWidth;
     Window	      transientFor;
     Window	      clientLeader;
     XSizeHints	      sizeHints;
     Pixmap	      pixmap;
-    CompTexture       texture;
+    CompTexture       *texture;
     CompMatrix        matrix;
     Damage	      damage;
     Bool	      inputHint;
@@ -1737,7 +1879,10 @@ struct _CompWindow {
     Bool shaded;
     Bool hidden;
 
+    unsigned int desktop;
+
     int pendingUnmaps;
+    int pendingMaps;
 
     char *startupId;
     char *resName;
@@ -1793,6 +1938,8 @@ struct _CompWindow {
     int      indexSize;
     int      vCount;
     int      texUnits;
+    int      texCoordSize;
+    int      indexCount;
 
     CompPrivate *privates;
 };
@@ -1841,6 +1988,18 @@ unsigned int
 getProtocols (CompDisplay *display,
 	      Window      id);
 
+unsigned int
+getWindowProp (CompDisplay  *display,
+	       Window	    id,
+	       Atom	    property,
+	       unsigned int defaultValue);
+
+void
+setWindowProp (CompDisplay  *display,
+	       Window       id,
+	       Atom	    property,
+	       unsigned int value);
+
 unsigned short
 getWindowProp32 (CompDisplay	*display,
 		 Window		id,
@@ -1924,6 +2083,11 @@ moveWindow (CompWindow *w,
 	    Bool       damage,
 	    Bool       immediate);
 
+void
+configureXWindow (CompWindow	 *w,
+		  unsigned int	 valueMask,
+		  XWindowChanges *xwc);
+
 void
 moveResizeWindow (CompWindow     *w,
 		  XWindowChanges *xwc,
@@ -1994,6 +2158,9 @@ windowGrabNotify (CompWindow   *w,
 void
 windowUngrabNotify (CompWindow *w);
 
+void
+windowStateChangeNotify (CompWindow *w);
+
 void
 moveInputFocusToWindow (CompWindow *w);
 
@@ -2085,6 +2252,13 @@ freeWindowIcons (CompWindow *w);
 int
 outputDeviceForWindow (CompWindow *w);
 
+Bool
+onCurrentDesktop (CompWindow *w);
+
+void
+setDesktopForWindow (CompWindow   *w,
+		     unsigned int desktop);
+
 
 /* plugin.c */
 
@@ -2096,14 +2270,19 @@ typedef void (*FiniPluginProc) (CompPlugin *plugin);
 
 typedef enum {
     CompPluginRuleBefore,
-    CompPluginRuleAfter
+    CompPluginRuleAfter,
+    CompPluginRuleRequire
 } CompPluginRule;
 
 typedef struct _CompPluginDep {
     CompPluginRule rule;
-    char	   *plugin;
+    char	   *name;
 } CompPluginDep;
 
+typedef struct _CompPluginFeature {
+    char *name;
+} CompPluginFeature;
+
 typedef struct _CompPluginVTable {
     char *name;
     char *shortDesc;
@@ -2130,6 +2309,9 @@ typedef struct _CompPluginVTable {
 
     CompPluginDep *deps;
     int		  nDeps;
+
+    CompPluginFeature *features;
+    int		      nFeatures;
 } CompPluginVTable;
 
 typedef CompPluginVTable *(*PluginGetInfoProc) (void);
@@ -2184,6 +2366,9 @@ windowFiniPlugins (CompWindow *w);
 CompPlugin *
 findActivePlugin (char *name);
 
+CompPluginFeature *
+findActiveFeature (char *name);
+
 CompPlugin *
 loadPlugin (char *plugin);
 
