@@ -480,6 +480,9 @@ isSwitchWin (CompWindow *w)
     if (!(ss->wMask & w->type))
 	return FALSE;
 
+    if (w->wmType & (CompWindowTypeDockMask | CompWindowTypeDesktopMask))
+	return FALSE;
+
     if (w->state & CompWindowStateSkipTaskbarMask)
 	return FALSE;
 
@@ -542,6 +545,8 @@ static void
 switchUpdateWindowList (CompScreen *s,
 			int	   count)
 {
+    int x, y;
+
     SWITCH_SCREEN (s);
 
     if (count > 1)
@@ -556,10 +561,15 @@ switchUpdateWindowList (CompScreen *s,
 
     ss->selectedWindow = ss->windows[0]->id;
 
+    x = s->outputDev[s->currentOutputDev].region.extents.x1 +
+	s->outputDev[s->currentOutputDev].width / 2;
+    y = s->outputDev[s->currentOutputDev].region.extents.y1 +
+	s->outputDev[s->currentOutputDev].height / 2;
+
     if (ss->popupWindow)
 	XMoveResizeWindow (s->display->display, ss->popupWindow,
-			   s->width  / 2 - WINDOW_WIDTH (count) / 2,
-			   s->height / 2 - WINDOW_HEIGHT / 2,
+			   x - WINDOW_WIDTH (count) / 2,
+			   y - WINDOW_HEIGHT / 2,
 			   WINDOW_WIDTH (count),
 			   WINDOW_HEIGHT);
 }
@@ -798,6 +808,10 @@ switchInitiate (CompScreen *s,
 			 XInternAtom (dpy, "_NET_WM_STATE", 0),
 			 XA_ATOM, 32, PropModeReplace,
 			 (unsigned char *) state, nState);
+
+	setWindowProp (s->display, ss->popupWindow,
+		       s->display->winDesktopAtom,
+		       0xffffffff);
     }
 
     if (!ss->grabIndex)
@@ -1403,9 +1417,8 @@ switchPaintScreen (CompScreen		   *s,
 	    switcher->destroyed = saveDestroyed;
 
 	    glPushMatrix ();
-	    glTranslatef (-0.5f, -0.5f, -DEFAULT_Z_CAMERA);
-	    glScalef (1.0f / s->width, -1.0f / s->height, 1.0f);
-	    glTranslatef (0.0f, -s->height, 0.0f);
+
+	    prepareXCoords (s, output, -DEFAULT_Z_CAMERA);
 
 	    if (!switcher->destroyed			 &&
 		switcher->attrib.map_state == IsViewable &&
