@@ -45,23 +45,40 @@ void
 donePaintScreen (CompScreen *screen) {}
 
 void
-translateRotateScreen (const ScreenPaintAttrib *sa)
+applyScreenTransform (CompScreen	      *screen,
+		      const ScreenPaintAttrib *sAttrib,
+		      int		      output)
 {
-    glTranslatef (sa->xTranslate,
-		  sa->yTranslate,
-		  sa->zTranslate + sa->zCamera);
+    glTranslatef (sAttrib->xTranslate,
+		  sAttrib->yTranslate,
+		  sAttrib->zTranslate + sAttrib->zCamera);
 
-    glRotatef (sa->xRotate, 0.0f, 1.0f, 0.0f);
-    glRotatef (sa->vRotate,
-	       1.0f - sa->xRotate / 90.0f,
+    glRotatef (sAttrib->xRotate, 0.0f, 1.0f, 0.0f);
+    glRotatef (sAttrib->vRotate,
+	       1.0f - sAttrib->xRotate / 90.0f,
 	       0.0f,
-	       sa->xRotate / 90.0f);
-    glRotatef (sa->yRotate, 0.0f, 1.0f, 0.0f);
+	       sAttrib->xRotate / 90.0f);
+    glRotatef (sAttrib->yRotate, 0.0f, 1.0f, 0.0f);
+}
+
+void
+prepareXCoords (CompScreen *screen,
+		int	   output,
+		float      z)
+{
+    glTranslatef (-0.5f, -0.5f, z);
+    glScalef (1.0f  / screen->outputDev[output].width,
+	      -1.0f / screen->outputDev[output].height,
+	      1.0f);
+    glTranslatef (-screen->outputDev[output].region.extents.x1,
+		  -screen->outputDev[output].region.extents.y2,
+		  0.0f);
 }
 
 void
 paintTransformedScreen (CompScreen		*screen,
 			const ScreenPaintAttrib *sAttrib,
+			Region			region,
 			int			output,
 			unsigned int		mask)
 {
@@ -70,17 +87,15 @@ paintTransformedScreen (CompScreen		*screen,
     int	       backgroundMask;
 
     if (mask & PAINT_SCREEN_CLEAR_MASK)
-	glClear (GL_COLOR_BUFFER_BIT);
+	clearTargetOutput (screen->display, GL_COLOR_BUFFER_BIT);
 
     screenLighting (screen, TRUE);
 
     glPushMatrix ();
 
-    translateRotateScreen (sAttrib);
+    (screen->applyScreenTransform) (screen, sAttrib, output);
 
-    glTranslatef (-0.5f, -0.5f, -sAttrib->zTranslate);
-    glScalef (1.0f / screen->width, -1.0f / screen->height, 1.0f);
-    glTranslatef (0.0f, -screen->height, 0.0f);
+    prepareXCoords (screen, output, -sAttrib->zTranslate);
 
     if (mask & PAINT_SCREEN_TRANSFORMED_MASK)
     {
@@ -91,8 +106,7 @@ paintTransformedScreen (CompScreen		*screen,
 	{
 	    backgroundMask |= PAINT_BACKGROUND_WITH_STENCIL_MASK;
 
-	    (*screen->paintBackground) (screen, &screen->region,
-					backgroundMask);
+	    (*screen->paintBackground) (screen, region, backgroundMask);
 
 	    glEnable (GL_STENCIL_TEST);
 
@@ -107,8 +121,7 @@ paintTransformedScreen (CompScreen		*screen,
 			continue;
 		}
 
-		(*screen->paintWindow) (w, &w->paint, &screen->region,
-					windowMask);
+		(*screen->paintWindow) (w, &w->paint, region, windowMask);
 	    }
 
 	    glDisable (GL_STENCIL_TEST);
@@ -121,7 +134,7 @@ paintTransformedScreen (CompScreen		*screen,
     else
 	windowMask = backgroundMask = 0;
 
-    (*screen->paintBackground) (screen, &screen->region, backgroundMask);
+    (*screen->paintBackground) (screen, region, backgroundMask);
 
     for (w = screen->windows; w; w = w->next)
     {
@@ -134,7 +147,7 @@ paintTransformedScreen (CompScreen		*screen,
 		continue;
 	}
 
-	(*screen->paintWindow) (w, &w->paint, &screen->region, windowMask);
+	(*screen->paintWindow) (w, &w->paint, region, windowMask);
     }
 
     glPopMatrix ();
@@ -156,8 +169,10 @@ paintScreen (CompScreen		     *screen,
 	{
 	    if (mask & PAINT_SCREEN_FULL_MASK)
 	    {
-		(*screen->paintTransformedScreen) (screen, sAttrib, output,
-						   mask);
+		region = &screen->outputDev[output].region;
+
+		(*screen->paintTransformedScreen) (screen, sAttrib, region,
+						   output, mask);
 
 		return TRUE;
 	    }
@@ -169,7 +184,9 @@ paintScreen (CompScreen		     *screen,
     }
     else if (mask & PAINT_SCREEN_FULL_MASK)
     {
-	(*screen->paintTransformedScreen) (screen, sAttrib, output, mask);
+	(*screen->paintTransformedScreen) (screen, sAttrib,
+					   &screen->outputDev[output].region,
+					   output, mask);
 
 	return TRUE;
     }
@@ -180,9 +197,7 @@ paintScreen (CompScreen		     *screen,
 
     glPushMatrix ();
 
-    glTranslatef (-0.5f, -0.5f, -DEFAULT_Z_CAMERA);
-    glScalef (1.0f / screen->width, -1.0f / screen->height, 1.0f);
-    glTranslatef (0.0f, -screen->height, 0.0f);
+    prepareXCoords (screen, output, -DEFAULT_Z_CAMERA);
 
     if (mask & PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK)
     {
@@ -408,7 +423,7 @@ addWindowGeometry (CompWindow *w,
 
 	for (it = 0; it < nMatrix; it++)
 	{
-	    if (matrix[it].xy != 0.0f && matrix[it].yx != 0.0f)
+	    if (matrix[it].xy != 0.0f || matrix[it].yx != 0.0f)
 	    {
 		rect = FALSE;
 		break;
@@ -823,7 +838,7 @@ paintWindow (CompWindow		     *w,
 	    mask |= PAINT_WINDOW_SOLID_MASK;
     }
 
-    if (!w->texture.pixmap)
+    if (!w->texture->pixmap)
     {
 	bindWindow (w);
 	if (!w->mapNum)
@@ -835,7 +850,7 @@ paintWindow (CompWindow		     *w,
     if (w->vCount)
     {
 	w->lastPaint = *attrib;
-	(*w->screen->drawWindowTexture) (w, &w->texture, attrib, mask);
+	(*w->screen->drawWindowTexture) (w, w->texture, attrib, mask);
     }
 
     return TRUE;
